#include <stdio.h>
#define __USE_LARGEFILE64
#define __USE_FILE_OFFSET64
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include "rt_parser.h"
#include "dtvstream.h"

extern int errno;

char input_file[MAX_NAME_LENGTH] = "";
char output_file[MAX_NAME_LENGTH] = "";
char output_file_name[MAX_NAME_LENGTH - MAX_EXTENSION_LENGTH];
char output_file_ext[MAX_EXTENSION_LENGTH];	// does not include "."
int in_fd;
int in_blocksize;
char in_open = 0;
unsigned long long buffer_file_offset;
unsigned long long lost_sync_offset;
int do_read_init;
char write_replace = 0;
char debug_output = 0;
char verbose_output = 0;
char info_only = 0;
#ifdef CHANNEL_TUNING
int channel_number = 0;
#endif
int program_number = 0;
int minutes_limit = 0;
int seconds_limit = 0;
int total_limit;
unsigned long long packet_limit = 0;
int max_file_segments = 0;
int max_file_count = 0;
char use_default_program = 0;
char use_all_packets = 0;
unsigned long long scan_count = DEFAULT_SCAN_COUNT;
char scan_all = 0;
int buffer_k_bytes = 0;
unsigned int selected_streams[MAX_SELECTED_STREAMS];		
int selected_streams_count;
char too_many_streams;
int input_is_char_dev;
unsigned long long write_byte_count;
char show_status;

unsigned char *input_buffer;
unsigned char *input_ptr;
unsigned char *input_end;

char input_eof;
int default_program = 0;

unsigned char pat_packet[TRANSPORT_PKT_SIZE];
unsigned char transport_stream_id[2];
int packet_counter = 0;

int got_interrupt;
int got_alarm;
int got_term;
unsigned long long processed_packet_count;
char alarm_setup;

long status_update_msecs = 5000;
int packet_resolution = 500;
struct timeval last_time;
unsigned long long next_count;
unsigned long long packet_interval;
char packet_interval_ok;
unsigned long long input_packet_count;
unsigned long long adj_packet_count;
char status_printed = 0;
unsigned long long read_packet_offset = 0;
unsigned long long seek_value;
unsigned long long seek_result;

static void sig_handler(int signo);
int pat_item_compare(const void *e1, const void *e2);
int scan_packets(void);
unsigned int calc_crc(unsigned char *buf, unsigned int len);
int process_packets(void);
int process_one_packet(unsigned char *packet);
int output_pat_packet (unsigned char *packet);
int get_new_packet(unsigned char **packet);
int fill_input_buffer(void);
int buffer_sync_ok(void);
void print_help(char *myname);
void set_output_file(char *fullname);
int convert_to_int(char *str, int *convert_error);
unsigned long long convert_to_int64(char *str, int *convert_error);
char * safe_strcat(char *dest, char *src, int src_copylen, int maxlen);
void close_input(void);
void add_stream(int stream);
int prompt_program(void);
void add_pmt_streams(struct queue_hdr *q_hdr);
void print_pmt(int prog_number, struct queue_hdr *q_hdr);
void get_channel_desc(int prog_number, char *channel_desc);
void get_lang(int prog_number, int pid_id, char *lang);
void update_status(unsigned long long count, char *op_text);
void end_status(void);
void secs_to_text(char *text, time_t secs);
unsigned long long timeval_subtract (struct timeval *x, struct timeval *y_arg);

#ifdef CHANNEL_TUNING
extern int tune_to_channel(int tuner_fd, int channel_number);
#endif

// writer
extern int init_writer(void);
extern int output_packet(unsigned char *packet);
extern int had_writer_error(void);
extern void close_output(void);
extern void end_writer(void);

// parser
// header for a Queue

struct queue_hdr {
	struct   buf_hdr     *q_head_ptr   ; /* proc Q head             */
	struct   buf_hdr     *q_tail_ptr   ; /* proc Q tail             */
	int                  q_tot_bytes   ; /* total table size        */
	short                q_valid       ; /* 1 - table valid         */
					     /* 0 - table NOT valid     */
	short                q_ver         ; /* 1 - table valid         */
};
extern void process_packet(char *packet_ptr, char do_init);
extern void prt_summary(void);
extern struct pat_hdr pat_hdr;
extern struct vct_hdr current_tvct_hdr;
extern unsigned long int g_vec;
struct queue_hdr *eit_queue = 0;

int main(int argc, char *argv[])
{
	char *arg;
	char *myname;
	char in_opts = 1;
	char argchar;
	int convert_error;
	char *afterslash, *enddot;
	int len;
	int saved_errno;
	int result;
	mode_t open_flags;
	mode_t open_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	off_t seek_result;
	struct stat in_stat;
	char new_extension[MAX_EXTENSION_LENGTH];
	
	arg = *argv++;
	argc--;
	myname = arg;

	// default is to show status if on a tty
	show_status = isatty(fileno(stdout));
	
	while (argc > 0) {
		arg = *argv++;
		argc--;

		// process options
		if (strcmp(arg, "--help") == 0) {
			print_help(myname);
			return 0;
			}
		if ((arg[0] == '-') && in_opts) {
			// need exactly one letter
			if (arg[1] == 0) {
				fprintf(stderr, "Missing option after \"-\".\n");
				print_help(myname);
				return 1;
				}
			else if (arg[2] != 0) {
				fprintf(stderr, "Invalid option: \"%s\".\n", arg);
				print_help(myname);
				return 1;
				}
			// process options
			argchar = arg[1];
			switch(argchar) {
				// "--" ends option processing
				case '-':
						in_opts = 0;
						break;
						
				// options with arguments
				case 'a':
				case 'b':
#ifdef CHANNEL_TUNING
				case 'c':
#endif
				case 'i':
				case 'k':
				case 'l':
				case 'm':
				case 'n':
				case 'N':
				case 'o':
				case 'p':
				case 's':
						if (argc > 0) {
							arg = *argv++;
							argc--;
							switch(argchar) {
								case 'a':
									if (strcmp(arg, "*") == 0) {
										scan_all = 1;
										}
									else {
										scan_count = convert_to_int64(arg, &convert_error);
										if ((convert_error != 0) || (scan_count < 0)) {
											fprintf(stderr, "Invalid argument \"%s\" for -a.\n", arg);
											print_help(myname);
											return 1;
											}
										}
									break;
								case 'b':
									read_packet_offset = convert_to_int64(arg, &convert_error);
									if ((convert_error != 0) || (packet_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -b.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
#ifdef CHANNEL_TUNING
								case 'c':
									channel_number = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (channel_number < 2) || (channel_number > 83)) {
										fprintf(stderr, "Invalid argument \"%s\" for -c.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
#endif
								case 'i':
									input_file[0] = 0;
									safe_strcat(input_file, arg, -1, MAX_NAME_LENGTH);
									break;
								case 'k':
									buffer_k_bytes = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (buffer_k_bytes < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -k.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'l':
									packet_limit = convert_to_int64(arg, &convert_error);
									if ((convert_error != 0) || (packet_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -l.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'm':
									minutes_limit = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (minutes_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -m.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'n':
									max_file_segments = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (max_file_segments < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -n.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'N':
									max_file_count = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (max_file_count < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -N.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 'o':
									set_output_file(arg);
									break;
								case 'p':
									if (strcmp(arg, "default") == 0) {
										use_default_program = 1;
										break;
										}
									if (strcmp(arg, "all") == 0) {
										use_all_packets = 1;
										break;
										}
									program_number = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (program_number < 1)) {
										fprintf(stderr, "Invalid argument \"%s\" for -p.\n", arg);
										print_help(myname);
										return 1;
										}
									break;
								case 's':
									seconds_limit = convert_to_int(arg, &convert_error);
									if ((convert_error != 0) || (seconds_limit < 0)) {
										fprintf(stderr, "Invalid argument \"%s\" for -s.\n", arg);
										print_help(myname);
										return 1;
									break;
								default:
									break;
									}
								}
							}							
						else {
							fprintf(stderr, "Argument missing for -%c.\n", arg[1]);
							print_help(myname);
							return 1;
							}
						break;
						
				case 'q':
						info_only = 1;
						break;

				case 'r':
						write_replace = 0;
						break;

				case 'R':
						write_replace = 1;
						break;

				case 'd':
						debug_output = 1;
						break;
						
				case 'v':
						verbose_output = 1;
						break;
						
				case 'e':
						show_status = 1;
						break;
						
				case 'E':
						show_status = 0;
						break;
						
				case 'h':
						print_help(myname);
						return 0;
						break;
						
				default:
						fprintf(stderr, "Invalid option: \"%s\".\n", arg);
						print_help(myname);
						return 1;
						break;
				}

			}

		// error if an argument specified
		else {
			fprintf(stderr, "Invalid argument: \"%s\" -- only options are used.\n", arg);
			print_help(myname);
			return 1;
			}
		}		
		
	// debug implies verbose
	if (debug_output) verbose_output = 1;

	// set file names not specified
	// set default input
	if (input_file[0] == 0) {
		strcpy(input_file, DEFAULT_INPUT_FILE);
		}

	// default output file is input_name.m2t
	if (output_file[0] == 0) {
		// copy input filename up to last period (if any)
		// for /dev get input file name following last /
		if (strncmp(input_file, "/dev/", 5) == 0) {
			afterslash = strrchr(input_file, '/');
			if (afterslash == 0) {
				afterslash = input_file;
				}
			else {
				afterslash++;
				}
			}
		else {
			afterslash = input_file;
			}
		
		len = 0;
		// don't use input file name after last period
		enddot = strrchr(afterslash, '.');
		if (enddot != 0) {
			len = enddot - afterslash;
			}
		if (len <= 0) {
			len = strlen(afterslash);
			}
		output_file_name[0] = 0;
		safe_strcat(output_file, afterslash, len,
					MAX_NAME_LENGTH - strlen(OUTPUT_NAME_EXTENSION));
		safe_strcat(output_file, OUTPUT_NAME_EXTENSION, -1, MAX_NAME_LENGTH);
		set_output_file(output_file);
		}
	
	// if packet limit specified, name this output file part 1
	// unless only 1 file specified
	if ((packet_limit > 0) && (max_file_count != 1)) {
		strcpy(new_extension, "1");
		if (output_file_ext[0] != 0) {
			safe_strcat(new_extension, ".", -1, MAX_EXTENSION_LENGTH);
			safe_strcat(new_extension, output_file_ext, -1, MAX_EXTENSION_LENGTH);
			}
		strcpy(output_file, output_file_name);
		strcat(output_file, ".");
		strcat(output_file, new_extension);
		}

	// check input and output files are different
	if (!info_only) {
		if (strcmp(input_file, output_file) == 0) {
			fprintf(stderr, "Input and output files are both \"%s\"\n", input_file);
			return 1;
			}
		}

	// compute total time limit
	total_limit = seconds_limit + minutes_limit * 60;

	// for debugging, display files and options
	if (debug_output) {
		printf("Settings:\n");
		printf("    Input file = \"%s\", Replace output = %d\n",
			   input_file, write_replace);
		printf("    Output name = \"%s\", Output extension = \"%s\", Output file = \"%s\"\n",
			   output_file_name, output_file_ext, output_file);	
		printf("    Minutes limit = %d, Seconds limit = %d, Total seconds limit = %d\n",
			   minutes_limit, seconds_limit, total_limit);	
#ifdef CHANNEL_TUNING
		printf("    Channel = %d, Program = %d, Packet scan count = %llu, Scan all = %d\n",
			   channel_number, program_number, scan_count, scan_all);
#else
		printf("    Program = %d, Packet scan count = %llu, Scan all = %d\n",
			   program_number, scan_count, scan_all);
#endif
		printf("    Beginning offset = %llu, Packet limit = %llu, Max segments = %d, Max files = %d, Output K-bytes = %d\n",
			   read_packet_offset, packet_limit, max_file_segments, max_file_count, buffer_k_bytes);
		printf("    Debug = %d, Verbose = %d Show elapsed = %d\n",
			   debug_output, verbose_output, show_status);
		}

	// open input file
	open_flags = O_RDONLY;
#ifdef O_LARGEFILE
	open_flags |= O_LARGEFILE;
	if (debug_output) printf("Using O_LARGEFILE for open.\n");
#else
	if (debug_output) printf("Not using O_LARGEFILE for open.\n");
#endif
	in_fd = open(input_file, open_flags);
	if (in_fd == -1) {
		fprintf(stderr, "Unable to open \"%s\": %s\n", input_file, strerror(errno));
		return(errno);
		}
	in_open = 1;

	// determine the size of the input file
	// determine if the input file is a character device and get optimal blocksize
	// (i.e. our tuner)
	result = fstat(in_fd, &in_stat);
	if (result != 0) {
		fprintf(stderr, "Unable to stat \"%s\": %s\n", input_file, strerror(errno));
		close_input();
		return(errno);
		}
	input_is_char_dev = S_ISCHR(in_stat.st_mode);
	in_blocksize = in_stat.st_blksize;
	input_packet_count = in_stat.st_size / TRANSPORT_PKT_SIZE;

	// if a beginning offset was specified, adjust to start reading there
	if ((input_is_char_dev) && (read_packet_offset > 0)) {
			fprintf(stderr, "Warning: beginning offset cannot be used for a character device.\n");
			read_packet_offset = 0;
			}

	if (read_packet_offset > 0) {
		if (read_packet_offset >= input_packet_count) {
			fprintf(stderr, "Starting offset %llu equals or exceeds file size %llu\n",
					read_packet_offset, input_packet_count);
			close_input();
			return(2);
			}
		
		// do seek
		seek_value = read_packet_offset * TRANSPORT_PKT_SIZE;
		seek_result = lseek(in_fd, seek_value, SEEK_SET);
		if (seek_result != seek_value) {
			fprintf(stderr, "Lseek failed for \"%s\": %s\n", input_file, strerror(errno));
			close_input();
			return(errno);
			}
	
		// adjust packet count
		input_packet_count -= read_packet_offset;
		}

	// if a packet limit and file count were specified, input_packet_count can't
	// exceed the limit times the number of files
	if ((packet_limit > 0) && (max_file_count > 0)) {
		adj_packet_count = packet_limit * max_file_count;
		if (adj_packet_count < input_packet_count) {
			input_packet_count = adj_packet_count;
			}
		}

	if (debug_output) {
		printf("character device = %d, optimal input blocksize = %d\n",
				input_is_char_dev, in_blocksize);
		}
	if (in_blocksize < TRANSPORT_PKT_SIZE * 3) {
		in_blocksize = REPLACEMENT_INPUT_BLKSIZE;
		if (debug_output) {
			printf("changed input blocksize to %d\n", in_blocksize);
			}
		}
	if (buffer_k_bytes == 0) {
		if (input_is_char_dev) {
			buffer_k_bytes = DEFAULT_CHAR_OUTPUT_SIZE;
			}
		else {
			buffer_k_bytes = DEFAULT_STD_OUTPUT_SIZE;
			}
		if (debug_output) {
			if (buffer_k_bytes != 0) {
				printf("changed output K-bytes to %d\n", buffer_k_bytes);
				}
			}
		}


#ifdef CHANNEL_TUNING
	// if a channel number was specified, attempt to tune to it
	if (channel_number != 0) {
		if (!input_is_char_dev) {
			fprintf(stderr, "Input file does not permit changing channels- not a character device.\n");
			close_input();
			return 3;
			}
		result = tune_to_channel(in_fd, channel_number);
		if (result != 0) {
			fprintf(stderr, "Unable to tune \"%s\" to channel %d: %s\n",
					input_file, channel_number, strerror(errno));
			close_input();
			return 4;
			}
		else if (verbose_output) {
			printf("Tuned \"%s\" to channel %d.\n", input_file, channel_number);
			}
		}
#endif

	// allocate the input buffer
	input_buffer = (unsigned char *)malloc(in_blocksize);
	if (input_buffer == 0) {
		fprintf(stderr, "Unable to allocate input buffer of size %u\n", in_blocksize);
		close_input();
		return ENOMEM;
		}

	// init input buffer
	do_read_init = 1;

	// set up for interrupts
	got_interrupt = 0;
	if (signal(SIGINT, sig_handler) == SIG_ERR) {
		fprintf(stderr, "Error setting SIGINT handler\n");
		result = 1;
		goto cleanup;
		}

	got_term = 0;
	if (signal(SIGTERM, sig_handler) == SIG_ERR) {
		fprintf(stderr, "Error setting SIGTERM handler\n");
		result = 1;
		goto cleanup;
		}

	got_alarm = 0;
	alarm_setup = 0;

	// handle time limit
	// if all packets are being analyzed, the time limit
	// applies to scanning
	if ((total_limit > 0) && (scan_all) && (!use_all_packets)) {
		if (signal(SIGALRM, sig_handler) == SIG_ERR) {
			fprintf(stderr, "Error setting SIGALRM handler\n");
			result = 1;
			goto cleanup;
			}
		alarm_setup = 1;
		alarm(total_limit);
		}

	// if we're not processing all packets, scan for program information
	if (!use_all_packets) {
		result = scan_packets();

		// cancel the alarm if we set it
		if ((total_limit > 0) && (scan_all)) {
			alarm(0);
			}

		if (result != 0) {
			goto cleanup;
			}
		}

	// done if we got a signal
	if (got_interrupt || got_term || got_alarm) {
		result = 1;
		goto cleanup;
		}

	// done if only info requested
	if (info_only) {
		result = 0;
		goto cleanup;
		}

	if (verbose_output && use_all_packets) {		// all may be set by scan
		printf("Processing all packets.\n");
		}

	// need to re-open input file if did scan
	if (!in_open) {
		// open flags already set
		in_fd = open(input_file, open_flags);
		if (in_fd == -1) {
			fprintf(stderr, "Unable to open \"%s\": %s\n", input_file, strerror(errno));
			result = errno;
			goto cleanup;
			}
		in_open = 1;

		// also need to do seek again
		if (read_packet_offset > 0) {
			// do seek
			seek_value = read_packet_offset * TRANSPORT_PKT_SIZE;
			seek_result = lseek(in_fd, seek_value, SEEK_SET);
			if (seek_result != seek_value) {
				fprintf(stderr, "Lseek failed for \"%s\": %s\n", input_file, strerror(errno));
				result = errno;
				goto cleanup;
				}
			}

		// reset input buffer
		do_read_init = 1;
		}

	// process packets
	if (result == 0) {
		// process
		result = process_packets();
		}
		
cleanup:
	// clean up and return
	free(input_buffer);
	close_input();

	return result;
}

static void sig_handler(int signo)
{
	switch (signo) {
		case SIGINT:
			status_printed = 0;
			got_interrupt = 1;
			printf("\nTerminating due to user interrupt.\n");
			break;
			
		case SIGALRM:
			end_status();
			got_alarm = 1;
			printf("Terminating at end of time limit.\n");
			break;
		
		case SIGTERM:
			end_status();
			got_term = 1;
			printf("Terminating due to kill signal.\n");
			break;

		default:
			break;		
		}
}

int pat_item_compare(const void *e1, const void *e2)
{
	int v1 = ((struct pat_item *)e1)->program_number;
	int v2 = ((struct pat_item *)e2)->program_number;
	return (v1<v2) ? -1 : (v1>v2) ? 1 : 0;
}

int scan_packets(void)
{
unsigned char *packet;
int result;
char init;
unsigned long long count;
struct pat_item *item_ptr;
int i, streams;
char channel_desc[32];
int program_to_use = 0;
int prompt_result = 0;
unsigned int ts_pid;
unsigned int pmt_pid;
unsigned int crc;

transport_stream_id[0] = 0;
transport_stream_id[1] = 0;

count = 0;
init = 1;

// initialize status
next_count = 0;
packet_interval_ok = 0;
status_printed = 0;

result = get_new_packet(&packet);
while ((result == 0) && (!got_interrupt) && (!got_term)
       && (!got_alarm)) {
	if (show_status) update_status(count, "Analyzing");
	if (!scan_all) {
		if (count >= scan_count) break;
		}
	ts_pid = ((packet[1] & 0x1f) << 8) | (packet[2] & 0xff);
	if (ts_pid == 0) {	// save transport stream id
		transport_stream_id[0] = packet[8];
		transport_stream_id[1] = packet[9];
		}
	process_packet(packet, init);
	init = 0;
	result = get_new_packet(&packet);
	count++;
	}

end_status();

// close input file
if (in_open) {
	close(in_fd);
	in_open = 0;
	}

if (count == 0) {
	fprintf(stderr, "No valid packets found.\n");
	return 1;
	}
	
if ((!scan_all) && (count < scan_count)) {
	fprintf(stderr, "Only able to scan %llu packets instead of %llu\n",
			count, scan_count);
	}

if (verbose_output) {
	printf("Summary after scanning first %llu packets:\n", count);
	prt_summary();
	}
	
// Count number of valid programs, and set default program
streams = 0;
if (pat_hdr.num_items > 0) {
	item_ptr = pat_hdr.pat_item_ptr;
	// check PMT table exists for each stream
	for (i = 0; i < pat_hdr.num_items; i++) {
		if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
			streams++;
			if (default_program == 0) {
				default_program = item_ptr->program_number;
				}
			else {
				if (default_program > item_ptr->program_number)
					default_program = item_ptr->program_number;
				}			
			}
		item_ptr++;
		}
	}
	
if (verbose_output && (streams != pat_hdr.num_items)) {
	printf("Stream contains %d total programs, %d valid programs.\n",
		pat_hdr.num_items, streams);
	}
else if (streams > 0) {
	printf("Stream contains %d program(s).\n", streams);
	}
else {
	fprintf(stderr, "No valid programs found in stream.\n");
	}
if (streams == 0) return 2;

if (verbose_output) {
	printf("Default program is %d.\n", default_program);
	}

// if just one stream and the program was not specified,
// use that stream
// error if different stream was specified

if (streams == 1) {
	if (program_number == 0) {
		program_number = default_program;
		}
	else if (program_number != default_program) {
		fprintf(stderr, "Requested program %d not found in stream.\n", program_number);
		return 3;
		}
	}

// Sort the pat_items by program number
qsort(pat_hdr.pat_item_ptr, pat_hdr.num_items,
	  sizeof(struct pat_item), pat_item_compare);

// Print each program and streams
item_ptr = pat_hdr.pat_item_ptr;
for (i = 0; i < pat_hdr.num_items; i++, item_ptr++) {
	if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
		if (program_number != 0) {  // listing specified program
			if (item_ptr->program_number != program_number) {
				continue;
				}
			else {		// listing the selected program
				if (info_only) {
					printf("    Program %d:", item_ptr->program_number);
					}
				else {
					printf("    Using program %d:", item_ptr->program_number);
					}
				program_to_use = program_number;
				}
			}
		else {					// listing all programs
			printf("    Program %d:", item_ptr->program_number);
			}
		// note: get_channel_desc will set default_program to minor program 1
		get_channel_desc(item_ptr->program_number, channel_desc);
		if (channel_desc[0] == 0) {
			printf("\n");
			}
		else {
			printf(" %s:\n", channel_desc);
			}
		print_pmt(item_ptr->program_number, item_ptr->pmt_q_hdr_ptr);
		}
	}

if (program_to_use == 0) {
	if (program_number != 0) {
		fprintf(stderr, "Requested program %d not found in stream.\n", program_number);
		return 3;
		}
	if (info_only) return 0;
	if (use_default_program) {
		program_to_use = default_program;
		printf("Using default program %d.\n", program_to_use);
		}
	else {
	new_prompt:
		prompt_result = prompt_program();
		if (use_all_packets) return 0;
		if (prompt_result == 0) return 4;
		// check returned number is a program number
		item_ptr = pat_hdr.pat_item_ptr;
		for (i = 0; i < pat_hdr.num_items; i++) {
			if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
				if (item_ptr->program_number == prompt_result) {
					program_to_use = prompt_result;
					}
				}
			item_ptr++;
			}
		if (program_to_use == 0) {
			printf("%d is not a valid program number\n", prompt_result);
			goto new_prompt;		
			}
		}
	}

if (info_only) return 0;

if (verbose_output) {
	printf("Selecting streams for program %d.\n", program_to_use);
	}

// define streams to output
selected_streams_count = 0;
too_many_streams = 0;

// PAT stream is not added- implicitly processed

// add the streams for the selected program
item_ptr = pat_hdr.pat_item_ptr;
for (i = 0; i < pat_hdr.num_items; i++) {
	if ((item_ptr->pmt_q_hdr_ptr)->q_head_ptr != 0) {
		if (item_ptr->program_number == program_to_use) {
			pmt_pid = item_ptr->program_map_pid;
			add_stream(pmt_pid);
			add_pmt_streams(item_ptr->pmt_q_hdr_ptr);
			break;
			}
		}
	item_ptr++;
	}

if (debug_output) {
	printf("%d streams selected:\n", selected_streams_count);
	for (i = 0; i < selected_streams_count; i++) {
		printf("    %02x\n", selected_streams[i]);
		}
	}

if (too_many_streams) {
	fprintf(stderr, "Stream table is too small for selected streams.\n");
	return 5;
	}

// generate the new PAT packet with just the selected streams

//initialize packet with 0xff
memset(pat_packet, 0xff, TRANSPORT_PKT_SIZE);

// set fields
pat_packet[0] = 0x47;		// sync
pat_packet[1] = 0x40;		// PUSI etc, PID
pat_packet[2] = 0;			// rest of PID
pat_packet[3] = 0x10;		// adaption etc, counter
pat_packet[4] = 0;			// offset
pat_packet[5] = 0;			// table id
pat_packet[6] = 0xb0;		// fixed bits, len
pat_packet[7] = 13;			// rest of len
pat_packet[8] = transport_stream_id[0];
pat_packet[9] = transport_stream_id[1];
pat_packet[10] = 0xc1;		// reserved, version, current
pat_packet[11] = 0;			// section number
pat_packet[12] = 0;			// last section number
pat_packet[13] = (program_to_use & 0x0000ff00) >> 8;
pat_packet[14] = program_to_use & 0x000000ff;
pat_packet[15] = 0xe0 | ((pmt_pid & 0x00001f00) >> 8);
pat_packet[16] = pmt_pid & 0x000000ff;
// calculate and store CRC
crc = calc_crc(pat_packet+5, 12);
pat_packet[17] = (crc & 0xff000000) >> 24;
pat_packet[18] = (crc & 0x00ff0000) >> 16;
pat_packet[19] = (crc & 0x0000ff00) >> 8;
pat_packet[20] = crc & 0x000000ff;

return 0;
}

unsigned int calc_crc(unsigned char *buf, unsigned int len)
{
   int               j;
   unsigned int      length;
   unsigned char     data_bit, next_byte;
   unsigned long int shift_reg;

   shift_reg = 0xFFFFFFFF;
   length = len;

    while ( length-- > 0 )
    {
	 next_byte = *buf++;
	 for (j=7; j>=0; j--)
	 {
	    data_bit = (next_byte>>j)&0x01;

	    data_bit ^= ((shift_reg>>31)&0x01);
	    shift_reg = (shift_reg<<1);
	    if ( data_bit == 1 )
	    {
		  shift_reg = (shift_reg ^ g_vec) | 1 ;
	    }
	 }
    }
    return( shift_reg );
}

void add_stream(int stream)
{
int i;

// room for stream?
if (selected_streams_count == MAX_SELECTED_STREAMS) {
	too_many_streams = 1;
	return;
	}

// if stream is already present, don't add again
if (selected_streams_count > 0) {
	for (i = 0; i < selected_streams_count; i++) {
		if (selected_streams[i] == stream) {
			return;
			}
		}
	}

selected_streams[selected_streams_count++] = stream;
}

int prompt_program(void)
{
char input_line[256];
int convert_error, result;

retry:
	printf("Enter program number to process, \"all\", or \"quit\":\n");
	scanf("%s", input_line);

	if (strcmp(input_line, "all") == 0) {
		use_all_packets = 1;
		return 0;
		}

	if (strcmp(input_line, "quit") == 0) {
		return 0;
		}

	result = convert_to_int(input_line, &convert_error);
	if (convert_error) {
		printf("\"%s\" is not a valid number\n", input_line);
		goto retry;
		}
	return result;
}

void add_pmt_streams(struct queue_hdr *q_hdr)
{
    char *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int tbl_sect_len;
    unsigned int len_left, descr_len;
    struct buf_hdr *buf_hdr_ptr;
    unsigned int pid_id;
    unsigned int pcr_pid_id;	// Program Clock Reference stream

    buf_hdr_ptr = q_hdr->q_head_ptr;
    buf_ptr = buf_hdr_ptr->buf_ptr;

	pcr_pid_id = ((buf_ptr[8]&0x1F)<<8) | (buf_ptr[9]&0xFF);
	if (pcr_pid_id != 0x1fff) add_stream(pcr_pid_id);
	tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	while ( len_left > 0 ) {
		pid_id = ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF);
		add_stream(pid_id);
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;
		}
}

void print_pmt(int prog_number, struct queue_hdr *q_hdr)
{
    char *buf_ptr, *c_ptr, *descr_ptr;
    unsigned int tbl_sect_len;
    unsigned int len_left, descr_len;
    struct buf_hdr *buf_hdr_ptr;
    char stream_name[128];
    char stream_lang[32];
    int pid_id;

    buf_hdr_ptr = q_hdr->q_head_ptr;
	if (buf_hdr_ptr == 0) {
		fprintf(stderr, "\tNo PMT table exists!\n");
		return;
		}
    buf_ptr = buf_hdr_ptr->buf_ptr;

    tbl_sect_len = ((buf_ptr[1]&0x0F)<<8) | (buf_ptr[2]&0xFF);
	descr_len = ((buf_ptr[10]&0x0F)<<8) | (buf_ptr[11]&0xFF) ;
	descr_ptr =  buf_ptr + 12;
	c_ptr = buf_ptr + descr_len + 12;
	len_left = (tbl_sect_len - 9 - 4) - (descr_len) ;
	while ( len_left > 0 )
	{
		pid_id = ((c_ptr[1]&0x1F)<<8) | (c_ptr[2]&0xFF);
		switch (c_ptr[0] & 0xff) {
			case 0x01:
					strcpy(stream_name, "Video (MPEG-1)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x02:
					strcpy(stream_name, "Video (MPEG-2)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x03:
					strcpy(stream_name, "Audio (MPEG-1)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x04:
					strcpy(stream_name, "Audio (MPEG-2)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			case 0x81:
					strcpy(stream_name, "Audio (Dolby AC-3)");
					get_lang(prog_number, pid_id,  stream_lang);
					if (stream_lang[0] != 0) {
						sprintf(stream_name + strlen(stream_name),
							", language = %s", stream_lang);
						}
					break;
			default:
					sprintf(stream_name, "Unknown (0x%02x)", c_ptr[0] & 0xff);
					break;
			}
		
	    printf("        %s\n", stream_name);
	    descr_len = ((c_ptr[3]&0x0F)<<8) | (c_ptr[4]&0xFF);
	    descr_ptr = c_ptr+5;
	    c_ptr += (descr_len + 5) ;
	    len_left -= (descr_len + 5) ;
	}
}


void get_channel_desc(int prog_number, char *channel_desc)
{
	int i, j;
	struct vct_hdr *current_vct_hdr;
	struct vct_item *vct_item_ptr;
	unsigned char short_name[8];
	
	channel_desc[0] = 0;
	
	current_vct_hdr = &current_tvct_hdr;
	vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
	if (vct_item_ptr == 0) return;
	for (i = 0; i < current_vct_hdr->num_entries; i++, vct_item_ptr++) {
		if (vct_item_ptr->program_number == prog_number) {
			for (j = 0; j < 7; j++) {
				short_name[j] = vct_item_ptr->short_name[j];
				}
			short_name[7] = 0;			
			sprintf(channel_desc, "%d-%d %s",
				vct_item_ptr->major_channel_number,
				vct_item_ptr->minor_channel_number,
				short_name);
			if (vct_item_ptr->minor_channel_number == 1) {
				default_program = prog_number;
				}
			break;
			}		
		}
}

void get_lang(int prog_number, int pid_id, char *lang)
{
	int i, j, k;
	struct vct_hdr *current_vct_hdr;
	struct vct_item *vct_item_ptr;
	unsigned char lang_name[4];
	
	lang[0] = 0;
	current_vct_hdr = &current_tvct_hdr;
	vct_item_ptr = current_vct_hdr->vct_tbl_ptr;
	if (vct_item_ptr == 0) return;
	for (i = 0; i < current_vct_hdr->num_entries; i++, vct_item_ptr++) {
		if (vct_item_ptr->program_number == prog_number) {
			for (k=0; k < vct_item_ptr->num_ch_elts; k++) {
				if (vct_item_ptr->elts[k].pes_pid == pid_id) {
					for (j = 0; j < 3; j++) {
						lang_name[j] = vct_item_ptr->elts[k].code_639[j];
						}
					lang_name[3] = 0;
					strcpy(lang, lang_name);				
					break;
					}
				}
			break;
			}		
		}
}


int process_packets(void)
{
unsigned char *packet;
int read_result, write_result;
static char init = 1;
static int count = 0;
struct pat_item *item_ptr;
int i, streams;
char channel_desc[32];
int program_to_use = 0;
int prompt_result = 0;
time_t start_time, elapsed_time;
int rate, hours, mins, secs;

write_result = init_writer();
if (write_result != 0) return write_result;
					 	   
// set up alarm if needed
if (total_limit > 0) {
	if (!alarm_setup) {
		if (signal(SIGALRM, sig_handler) == SIG_ERR) {
			fprintf(stderr, "Error setting SIGALRM handler\n");
			return 1;
			}
		}
	if (!input_is_char_dev) {
		fprintf(stderr, "Warning: Setting processing time limit when input is not a device.\n");
		}
	alarm(total_limit);
	}
	
processed_packet_count = 0;
write_byte_count = 0;

// initialize status
next_count = 0;
packet_interval_ok = 0;
status_printed = 0;

start_time = time(0);
read_result = get_new_packet(&packet);
while ((read_result == 0) && (write_result == 0) 
       && (got_interrupt == 0) && (got_alarm == 0)
       && (got_term == 0) && (!had_writer_error())) {
	if (show_status) {
		if (input_is_char_dev) {
			update_status(processed_packet_count, "Recording");	
			}
		else {
			update_status(processed_packet_count, "Processing");
			}
		}
	processed_packet_count++;
	write_result = process_one_packet(packet);
	if (write_result == 0) {
		read_result = get_new_packet(&packet);
		}
	}

if (write_result == REACHED_FILE_LIMIT) {		// simulate eof if reached file count limit
	processed_packet_count--;
	write_result = 0;
	input_eof = 1;
	}

end_writer();

if (show_status) end_status();

elapsed_time = time(0) - start_time;
if (elapsed_time > 0) {
	rate = processed_packet_count / elapsed_time;
	}
else {
	rate = 0;
	}

hours = elapsed_time / 3600;
elapsed_time -= hours * 3600;
mins = elapsed_time / 60;
elapsed_time -= mins * 60;
secs = elapsed_time;

printf("Processed %llu packets, Wrote %llu packets.\n",
	processed_packet_count, write_byte_count / TRANSPORT_PKT_SIZE);
	
printf("Elapsed time = %02d:%02d:%02d, Rate = %d packets/second.\n",
	hours, mins, secs, rate);

if (input_eof) {
	read_result = 0;
	}

if (write_result != 0) {
	return write_result;
	}
else {
	return read_result;
	}
}

int process_one_packet(unsigned char *packet)
{
unsigned int ts_pid;
int i, result;

result = 0;

// if doing selected packets, check if this one should be included
if (use_all_packets) {
	result = output_packet(packet);
	}
else {
	ts_pid = ((packet[1] & 0x1f) << 8) | (packet[2] & 0xff);
	if (ts_pid == 0) result = output_pat_packet(packet);
	else {
		for (i = 0; i < selected_streams_count; i++) {
			if (selected_streams[i] == ts_pid) {
				result = output_packet(packet);
				break;
				}		
			}
		}
	}

return result;
}

int output_pat_packet (unsigned char *packet)
{
	int current, section;
	
	// skip packets for additional sections
	section = packet[11];
	if (section != 0) return 0;
	
	// skip packet which is not current
	current = packet[10] & 0x01;
	if (current == 0) return 0;

	// store counter value in packet
	pat_packet[3] = 0x10 | (packet_counter & 0x0f);
	packet_counter++;
	if (packet_counter == 16) packet_counter = 0;

	// output our PAT packet
	return(output_packet(pat_packet));
}


int get_new_packet(unsigned char **packet)
{
int remain;
int result;
unsigned int sync_count;

if (do_read_init) {
	input_ptr = input_buffer;
	input_end = input_buffer;
	input_eof = 0;
	remain = 0;
	}
else {
	// get remaining length in buffer
	remain = input_end - input_ptr;
	remain -= TRANSPORT_PKT_SIZE;			// did this packet
	}	
	
// if not long enough, refill buffer
if (remain < TRANSPORT_PKT_SIZE) {
	// return done if last read reached eof
	if (input_eof) return -1;

	// move any remaining data to the start of the buffer
	if (remain > 0) {
		input_ptr += TRANSPORT_PKT_SIZE;
		memmove(input_buffer, input_ptr, remain);
		input_end = input_buffer + remain;
		}
	else {
		input_end = input_buffer;
		}
	if (do_read_init) {
		buffer_file_offset = 0;		// first read
		do_read_init = 0;
		}
	else {
		buffer_file_offset += in_blocksize - remain;
		}
	result = fill_input_buffer();
	if (result != 0) {
		return result;
		}
	remain = input_end - input_buffer;
	if (remain < TRANSPORT_PKT_SIZE) {
		return -1;
		}
	input_ptr = input_buffer;
	}
else {
	input_ptr += TRANSPORT_PKT_SIZE;
	}	

// if packet is valid, ready to return
if (*input_ptr == SYNC_BYTE) {
	*packet = input_ptr;
	return 0;
	}

// need to resynchronize
lost_sync_offset = buffer_file_offset + (input_ptr - input_buffer);
if ((lost_sync_offset != 0) || (verbose_output)) {
	fprintf(stderr, "Lost sync at input offset %llu, output packet count = %llu.\n",
			lost_sync_offset,  write_byte_count / TRANSPORT_PKT_SIZE);
	}

sync_count = 0;
while ((!buffer_sync_ok()) && (sync_count < SYNC_SEARCH_BYTE_COUNT)) {
	// at this point, input_ptr is at the start of one or more full
	// packets, but the starting point is out of sync

	// try to move input_ptr ahead to another sync byte
	remain = input_end - input_ptr - 1;
	input_ptr = memchr(input_ptr + 1, SYNC_BYTE, remain);
	sync_count += remain;
	if (input_ptr == 0) input_ptr = input_end;
	remain = input_end - input_ptr;		// 0 if sync not found
	// try to get at least 3 packets to check
	if (remain < (TRANSPORT_PKT_SIZE * 3)) {
		// discard data before possible sync byte
		memmove(input_buffer, input_ptr, remain);
		input_ptr = input_buffer;
		input_end = input_buffer + remain;
		buffer_file_offset += in_blocksize - remain;
		// try to fill buffer 
		if (!input_eof) {
			result = fill_input_buffer();
			if (result != 0) {
				return result;
				}
			}
		remain = input_end - input_buffer;
		// handle as eof if not at least 3 packets to check
		if (remain < (TRANSPORT_PKT_SIZE * 3)) {
			return -1;
			}
		}
	}

if (buffer_sync_ok()) {
	if ((lost_sync_offset != 0) || (verbose_output)) {
		fprintf(stderr, "Achieved sync after %d bytes.\n",
				(int)((buffer_file_offset + (input_ptr - input_buffer)) - lost_sync_offset));
		}
	*packet = input_ptr;
	return 0;
	}
else {
	fprintf(stderr, "Failed to achieve sync after %uK bytes.\n",
			(sync_count + 1023) / 1024);
	return 1;
	}
}

int fill_input_buffer(void)
{
ssize_t read_length, read_actual;


// fill input buffer
// assumes input_eof has been checked
// and input_end is set to end of data to keep

// fill the buffer from the file
read_length = in_blocksize - (input_end - input_buffer);
while (read_length > 0) {
	read_actual = read(in_fd, input_end, read_length);
	if (read_actual < 0) {
		#ifdef EINTR
		if (errno == EINTR) {
			continue;
			}
		#endif
		fprintf(stderr, "Error reading input file %s: %s\n", input_file, strerror(errno));
		return errno;
		}
	if (read_actual == 0) {
		input_eof = 1;
		return 0;
		}
	input_end += read_actual;
	read_length = in_blocksize - (input_end - input_buffer);
	}
	
return 0;
}

int buffer_sync_ok(void)
{
	int remain;
	
	remain = input_end - input_ptr;
	
	if (remain >= (TRANSPORT_PKT_SIZE * 3)) {
		if ((*(input_ptr + (TRANSPORT_PKT_SIZE * 2))) != SYNC_BYTE){
			return 0;
			} 
		}

	if (remain >= (TRANSPORT_PKT_SIZE * 2)) {
		if ((*(input_ptr + TRANSPORT_PKT_SIZE)) != SYNC_BYTE){
			return 0;
			} 
		}

	if (remain >= TRANSPORT_PKT_SIZE) {
		return ((*input_ptr) == SYNC_BYTE);
		}

	return 0;
}

void print_help(char *myname)
{
	fprintf(stderr, "dtvstream version %s\n", VERSION);
	fprintf(stderr, "Format: %s [options]\n", myname); 
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "    -i input-file   Defaults to /dev/dtv\n");
	fprintf(stderr, "    -o output-file  Defaults to input-name.m2t\n");
#ifdef CHANNEL_TUNING
	fprintf(stderr, "    -c channel      tune to indicated TV channel (2 - 83)\n");
#endif
	fprintf(stderr, "    -p prog         select the given DTV program number\n");
	fprintf(stderr, "                    \"default\" for the default program\n");
	fprintf(stderr, "                    \"all\" to process all packets\n");
	fprintf(stderr, "    -m mins         stop after this many minutes (added to secs)\n");
	fprintf(stderr, "    -s secs         stop after this many seconds (added to mins)\n");
	fprintf(stderr, "    -b offset       beginning packet offset for reading from file\n");
	fprintf(stderr, "    -l count        start a new output file after every count %d-byte packets\n", TRANSPORT_PKT_SIZE);
	fprintf(stderr, "    -n count        number of last file to create (gets all remaining packets)\n");
	fprintf(stderr, "    -N count        stop after writing count files\n");
	fprintf(stderr, "    -a count or *   number of initial packets to analyze (default %d)\n", DEFAULT_SCAN_COUNT); 
	fprintf(stderr, "    -k size         maximum output buffer space (use nnn for nnnK)\n");
	fprintf(stderr, "                    default %d for tuner (unlimited), %d for other files\n", DEFAULT_CHAR_OUTPUT_SIZE, DEFAULT_STD_OUTPUT_SIZE);
	fprintf(stderr, "    -e    show elapsed/remaing time (default for terminal)\n");
	fprintf(stderr, "    -E    don't show elapsed/remaining time (default for file)\n");
	fprintf(stderr, "    -r    don't relace existing output file (default)\n");
	fprintf(stderr, "    -R    replace existing output file\n");
	fprintf(stderr, "    -q    just display stream program information\n");
	fprintf(stderr, "    -v    enable verbose output\n");
	fprintf(stderr, "    -d    enable debugging output (implies -v)\n");
	fprintf(stderr, "    -h    print this help text (also --help)\n");
}

int convert_to_int(char *str, int *convert_error)
{
	int result;
	char *strptr;
	
	result = (int)strtol(str, &strptr, 10);
	if (convert_error == 0) return result;
	
	if (*strptr == 0) {
		*convert_error = 0;
		}
	else {
		*convert_error = 1;
		}

	return result;
}

unsigned long long convert_to_int64(char *str, int *convert_error)
{
	unsigned long long result;
	int i, count;

	if (str == 0) {
		*convert_error = 1;
		return 0;
		}

	if ((*str) == 0) {
		*convert_error = 1;
		return 0;
		}

	for (i = 0; i < strlen(str); i++) {
		if (!isdigit(str[i])) {
			*convert_error = 1;
			return 0;
			}
		}

	result = 0;
	count = sscanf(str, "%llu", &result);
	*convert_error = ((count != 1) || (errno != 0));
	return result;
}

void set_output_file(char *fullname)
{
	char *enddot;
	int extlen;

	// look for final "." in name
	enddot = strrchr(fullname, '.');
	extlen = 0;
	if (enddot != 0) {
		extlen  = (fullname + strlen(fullname)) - enddot;
		}

	if (extlen < 2) {			// no period or just period	at end	
		output_file_name[0] = 0;
		safe_strcat(output_file_name, fullname, -1, MAX_NAME_LENGTH - MAX_EXTENSION_LENGTH);

		output_file_ext[0] = 0;

		strcpy(output_file, output_file_name);
		}

	else {						// non-null extension
		output_file_name[0] = 0;
		safe_strcat(output_file_name, fullname, enddot - fullname, MAX_NAME_LENGTH - MAX_EXTENSION_LENGTH);

		output_file_ext[0] = 0;
		safe_strcat(output_file_ext, enddot + 1, extlen - 1, MAX_EXTENSION_LENGTH); 
		
		strcpy(output_file, output_file_name);
		strcat(output_file, ".");
		strcat(output_file, output_file_ext);
		}
}

char * safe_strcat(char *dest, char *src, int src_copylen, int maxlen)
{
// len = declared length of dest
// strcat which appends only as much as fits
int src_len, dest_len, copylen;

if (src == 0) return dest;	// no src string
src_len = strlen(src);
if (src_len == 0) return dest;		// nothing to copy
if (src_copylen == 0) return dest;	// zero length requested 
if ((src_copylen > 0) && (src_copylen < src_len)) {
	src_len = src_copylen;			// use requested length if valid
	}

dest_len = strlen(dest);
copylen = maxlen - 1 - dest_len;
if (copylen <= 0) return dest; // already no room

if (copylen > src_len) copylen = src_len;  // no longer than source

memcpy(dest + dest_len, src, copylen);
dest[dest_len + copylen] = 0;

return dest;
}

void close_input(void)
{
if (in_open) {
	close(in_fd);
	in_open = 0;
	}
}

void update_status(unsigned long long count, char *op_text)
{
struct timeval current_time;
static struct timeval start_time;
struct timezone tz;
unsigned long long elapsed;
time_t remaining_time;
char elapsed_text[12];
char remaining_text[12];
int percent_done;

if (count != next_count) return;

// normal processing
if (packet_interval_ok) {
	// recompute packet interval based on current rate
	gettimeofday(&current_time, &tz);
	elapsed = (timeval_subtract(&current_time, &last_time) + 500) / 1000;
	if (elapsed == 0) {
		// should never happen- just a precaution
		// keep previous interval
		}
	else {
		packet_interval = (status_update_msecs * packet_interval) / elapsed;
		}
	goto do_status;
	return;
	}

// initializing
if (count == 0) {
	gettimeofday(&start_time, &tz);
	last_time = start_time;
	next_count = packet_resolution;
	return;
	}

// getting initial interval
gettimeofday(&current_time, &tz);
elapsed = (timeval_subtract(&current_time, &last_time) + 500) / 1000;
if (elapsed < status_update_msecs) {
	next_count += packet_resolution;
	return;
	}

// set initial interval
packet_interval_ok = 1;
packet_interval = (status_update_msecs * next_count) / elapsed;
printf("%s...\n", op_text);

do_status:

// update for next call
next_count += packet_interval;
last_time = current_time;

// output the status
// different depending on whether total packet count is known

// compute remaining time based on average rate from start
elapsed = (timeval_subtract(&current_time, &start_time) + 500) / 1000;
secs_to_text(elapsed_text, (elapsed + 500) / 1000);
if (input_packet_count == 0) {
	printf("  Elapsed time: %s\r", elapsed_text);
	}
else {
	if (elapsed == 0) {
		strcpy(remaining_text, "????");		// shouldn't happen
		}
	else {
		// get average rate
		remaining_time = ((input_packet_count - count) * elapsed) / count;
		secs_to_text(remaining_text, (remaining_time + 500) / 1000);
		}
	percent_done = (count * 100) / input_packet_count;
	printf("  Finished %d%%, elapsed = %s, remaining = %s          \r",
			percent_done, elapsed_text, remaining_text); 
	}
status_printed = 1;
fflush(stdout);
}

void end_status(void)
{
if (status_printed) {
	status_printed = 0;
	printf("\n");
	fflush(stdout);
	}
}

void secs_to_text(char *text, time_t total_secs)
{
	int hours, mins, secs;

	hours = total_secs / 3600;
	total_secs -= hours * 3600;
	mins = total_secs / 60;
	secs = total_secs - (mins * 60);
	
	if (hours > 0) {
		sprintf(text, "%d:%02d:%02d", hours, mins, secs);
		}
	else {
		sprintf(text, "%d:%02d", mins, secs);
		}
}

// Calulate timeval difference for x - y
// returns difference in microseconds (resolution system-dependent)

unsigned long long timeval_subtract (struct timeval *x, struct timeval *y_arg)
{
struct timeval y;
int nsec;
unsigned long long result;

y = *y_arg;	// copy y because we modify it

// Perform the carry for the later subtraction by updating y.
if (x->tv_usec < y.tv_usec) {
	nsec = (y.tv_usec - x->tv_usec) / 1000000 + 1;
	y.tv_usec -= 1000000 * nsec;
	y.tv_sec += nsec;
	}
if (x->tv_usec - y.tv_usec > 1000000) {
	nsec = (y.tv_usec - x->tv_usec) / 1000000;
	y.tv_usec += 1000000 * nsec;
	y.tv_sec -= nsec;
	}

// Compute difference-
// 'tv_usec' difference is certainly positive.
result = (x->tv_sec - y.tv_sec) * 1000000;
result += x->tv_usec - y.tv_usec;

return result;
}

