#include <stdio.h>
#include <stdarg.h>

#define MAX_MSG_LENGTH	400
static char msgbuff[MAX_MSG_LENGTH];
static char savebuff[MAX_MSG_LENGTH];

// external
int write_msg(char msgsource, char msgtype, char *format, ...);
void start_msg(char *format, ...);
void append_msg(char *format, ...);
int end_msg(char msgsource, char msgtype, char *format, ...);

// internal
int get_saved_length(void);
int print_string(char msgsource, char msgtype, char *string);

int write_msg(char msgsource, char msgtype, char *format, ...)
{
	va_list varargs;
	
	va_start(varargs, format);
	vsnprintf(msgbuff, MAX_MSG_LENGTH, format, varargs);
	va_end(varargs);

	return(print_string(msgsource, msgtype, msgbuff));
}

void start_msg(char *format, ...)
{
	va_list varargs;
	
	va_start(varargs, format);
	vsnprintf(savebuff, MAX_MSG_LENGTH, format, varargs);
	va_end(varargs);
}

void append_msg(char *format, ...)
{
	va_list varargs;
	int len;

	len = get_saved_length();

	if (len < 0) {	// must not have called start
		len = 0;
		}

	if (len == (MAX_MSG_LENGTH - 1)) {	// check if full
		return;
		}	

	va_start(varargs, format);
	vsnprintf(savebuff+len, MAX_MSG_LENGTH-len, format, varargs);
	va_end(varargs);
}

int end_msg(char msgsource, char msgtype, char *format, ...)
{
	va_list varargs;
	int len;

	len = get_saved_length();

	if (len < 0) {	// must not have called start or add
		len = 0;
		}

	if (len == (MAX_MSG_LENGTH - 1)) {	// check if full
		return;
		}	

	va_start(varargs, format);
	vsnprintf(savebuff+len, MAX_MSG_LENGTH-len, format, varargs);
	va_end(varargs);

	return(print_string(msgsource, msgtype, savebuff));
}

int get_saved_length(void)
{
	int i;

	for (i = 0; i < MAX_MSG_LENGTH; i++) {
		if (savebuff[i] == 0) return i;
		}

	return -1;
}

int print_string(char msgsource, char msgtype, char *string)
{
	FILE *outfile;
	char *format;
	char *nl_format = "%s\n";
	char *progress_format = "%s";
	int result;
	
	// get format to use
	if (msgtype == 'P') {
		format = progress_format;
		}
	else {
		format = nl_format;
		}

	// get file to use
	if (msgtype == 'E') {
		outfile = stderr;
		}
	else {
		outfile = stdout;
		}

	result = fprintf(outfile, format, string);
	fflush(outfile);
	return result;
}
