/*
 * parse ATSC (subset of MPEG-2) transport stream
 *
 *  difines pkt bit fields & some data structures for them
 *
 */


struct transport_pkt_hdr {
	unsigned int sync_byte                            : 8;
	unsigned int transport_error_indicator            : 1;
	unsigned int payload_unit_start_indicator         : 1;
	unsigned int transport_priority                   : 1;
	unsigned int pid                                  :13;
	unsigned int transport_scrambling_control         : 2;
	unsigned int adaptation_field_control             : 2;
	unsigned int contiuity_counter                    : 4;
	unsigned int adaptation_field_length              : 8;
	unsigned int discontinuity_indicator              : 1;
	unsigned int random_access_indicator              : 1;
	unsigned int elementary_stream_priority_indicator : 1;
	unsigned int PCR_flag                             : 1;
	unsigned int OPCR_flag                            : 1;
	unsigned int splicing_point_flag                  : 1;
	unsigned int transport_private_data_flag          : 1;
	unsigned int adaptation_field_extension_flag      : 1;
} ;

struct pes_pkt_hdr {
	unsigned int packet_start_code_prefix       : 24;
	unsigned int stream_id                      :  8;
	unsigned int pes_packet_length              : 16;
} ;

struct gen_tbl_hdr {
	unsigned int table_id                       :  8;
	unsigned int section_syntax_indicator       :  1;
	unsigned int private_indicator              :  1;
	unsigned int zero                           :  2;
	unsigned int section_length                 : 12;
	unsigned int various_id                     : 16;
	unsigned int reserved                       :  2;
	unsigned int version_number                 :  5;
	unsigned int current_next_indicator         :  1;
	unsigned int section_number                 :  8;
	unsigned int last_section_number            :  8;
	unsigned int protocol_version               :  8;
} ;

struct mgt_field_hdr {
	unsigned int table_id                       :  8;
	unsigned int section_syntax_indicator       :  1;
	unsigned int private_indicator              :  1;
	unsigned int zero                           :  2;
	unsigned int section_length                 : 12;
	unsigned int various_id                     : 16;
	unsigned int reserved                       :  2;
	unsigned int version_number                 :  5;
	unsigned int current_next_indicator         :  1;
	unsigned int section_number                 :  8;
	unsigned int last_section_number            :  8;
	unsigned int protocol_version               :  8;
	unsigned int tables_defined                 : 16;
} ;

struct vct_tbl_hdr {
	unsigned int table_id                       :  8;
	unsigned int section_syntax_indicator       :  1;
	unsigned int private_indicator              :  1;
	unsigned int zero                           :  2;
	unsigned int section_length                 : 12;
	unsigned int transport_stream_id            : 16;
	unsigned int reserved                       :  2;
	unsigned int version_number                 :  5;
	unsigned int current_next_indicator         :  1;
	unsigned int section_number                 :  8;
	unsigned int last_section_number            :  8;
	unsigned int protocol_version               :  8;
	unsigned int num_channels_in_section        :  8;
};

struct mgt_fields {
	unsigned int table_type                     : 16;
	unsigned int reserved_1                     :  3;
	unsigned int table_type_pid                 : 13;
	unsigned int reserved_2                     :  3;
	unsigned int table_type_version_number      :  5;
	unsigned int number_bytes                   : 32;
	unsigned int reserved_3                     :  4;
	unsigned int table_type_descriptor_length   : 12;
} ;

struct mgt_descr_struct {
	unsigned int reserved                       :  4;
	unsigned int descriptors_length             : 12;
	char         descr[2]                            ;
} ;

struct mgt_tbl_hdr {
	unsigned int mgt_version                        ;
	unsigned int mgt_num_entries                    ;
	struct       mgt_tbl_entry  *mgt_tbl_ptr        ;
	unsigned int descriptor_length                  ;
	char         *mgt_descriptor_ptr                ;
} ;

struct mgt_tbl_entry {
	unsigned int table_type                         ;
	unsigned int table_type_pid                     ;
	unsigned int table_type_version_number          ;
	unsigned int number_bytes                       ;
	struct       queue_hdr   *table_type_q_ptr      ;
	unsigned int table_type_descriptor_length       ;
	char         *table_type_descriptor_ptr         ;
/*      unsigned int mgt_descriptor_length              ;
 */
} ;

struct vct_fields {
   /*   unsigned int short_name[7]                  : 16;  */
	short        short_name[7]                      ;
	unsigned int reserved_1                     :  4;
	unsigned int major_channel_number           : 10;
	unsigned int minor_channel_number           : 10;
	unsigned int modulation_mode                :  8;
	unsigned int carrier_frequency              : 32;
	unsigned int channel_TSID                   : 16;
	unsigned int program_number                 : 16;
	unsigned int etm_location                   :  2;
	unsigned int access_controlled              :  1;
	unsigned int hidden                         :  1;
	unsigned int path_select                    :  1;
	unsigned int out_of_band                    :  1;
	unsigned int reserved_2                     :  4;
	unsigned int service_type                   :  6;
	unsigned int source_id                      : 16;
	unsigned int reserved_3                     :  6;
	unsigned int descriptor_length              : 10;
	char         *descriptor_ptr                    ;
} ;

struct vct_addl  {
	unsigned int reserved                       :  6;
	unsigned int additional_descriptors_length  : 10;
} ;

struct descr_struct {
	unsigned int tag                            :  8;
	unsigned int length                         :  8;
	char         descr[2]                            ;
} ;

struct vct_ch_descr_hdr  {
	unsigned int tag                            :  8;
	unsigned int length                         :  8;
	unsigned int reserved                       :  3;
	unsigned int pcr_pid                        : 13;
	unsigned int number_elements                :  8;
};

struct elt_descr  {
	unsigned int stream_type                    :  8;
	unsigned int reserved                       :  3;
	unsigned int pes_pid                        : 13;
   /*   char         iso_code_639[3]                :  8; */
	char         iso_code_639[3]                    ;
} ;

struct vct_hdr {
	unsigned int       version_number               ;
	unsigned int       transport_stream_id          ;
	unsigned int       num_entries                  ;
	unsigned int       descr_length                 ;
		 char      *descr_ptr                   ;
	struct   vct_item  *vct_tbl_ptr                 ;
} ;

struct ch_elts {
	unsigned int stream_id;
	unsigned int pes_pid;
	char         code_639[3];
};

struct vct_item {
	unsigned int     short_name[7]             ;
	unsigned int     major_channel_number      ;
	unsigned int     minor_channel_number      ;
	unsigned int     modulation_mode           ;
	unsigned int     carrier_frequency         ;
	unsigned int     channel_TSID              ;
	unsigned int     program_number            ;
	unsigned int     etm_location              ;
	unsigned char    access_controlled         ; /* the bits are there, */
	unsigned char    hidden                    ; /* just not used */
	unsigned char    path_select               ; /*  - only for cable */
	unsigned char    out_of_band               ; /*  - only for cable */
	unsigned int     service_type              ;
	unsigned int     source_id                 ;
	unsigned int     pcr_pid                   ;
	unsigned int     num_ch_elts               ;
	unsigned int     descr_length              ;
		 char    *descr_ptr                ;
	struct   ch_elts elts[10]                  ;
} ;

struct stt_fields {
	unsigned int system_time                    : 32;
	unsigned int gps_utc_offset                 :  8;
	unsigned int ds_status                      :  1;
	unsigned int reserved                       :  2;
	unsigned int ds_day_of_month                :  5;
	unsigned int ds_hour                        :  8;
} ;

struct stt_data {
	unsigned int  system_time                       ;
	unsigned int  gps_utc_offset                    ;
	unsigned int  ds_status                         ;
	unsigned int  ds_day_of_month                   ;
	unsigned int  ds_hour                           ;
	unsigned int  descriptor_length                 ;
	unsigned char *descriptor_ptr                   ;
} ;

struct eit_fields {
	unsigned int reserved_1                     :  2;
	unsigned int event_id                       : 14;
	unsigned int start_time                     : 32;
	unsigned int reserved_2                     :  2;
	unsigned int etm_location                   :  2;
	unsigned int length_in_seconds              : 20;
	unsigned int title_length                   :  8;
	char         *title_text_ptr                    ;
	unsigned int reserved_3                     :  4;
	unsigned int descriptors_length             : 12;
	char         *descriptor_ptr                    ;
} ;

struct eit_tbl_x_hdr {
	struct       eit_tbl_x_hdr  *next_eit_tbl_hdr   ;
	unsigned int version_number                     ;
	unsigned int section_number                     ;
	unsigned int last_section_number                ;
	char         *descriptor_ptr                    ;
} ;

struct eit_section {
	unsigned int section_number                     ;
	unsigned int num_events_in_section              ;
	struct       eit_section  *next_eit_section_ptr ;
	struct       eit_entry    *eit_section_tbl_ptr  ;
};

struct eit_entry {
	unsigned int event_id                   ;
	unsigned int start_time                 ;
	unsigned int etm_location               ;
	unsigned int length_in_seconds          ;
	unsigned int title_length               ;
	char         *title_text_ptr            ;
	unsigned int descriptors_length         ;
	char         *descriptor_ptr            ;
} ;

struct pat_field_hdr {
	unsigned int table_id                       :  8;
	unsigned int section_syntax_indicator       :  1;
	unsigned int zero                           :  1;
	unsigned int reserved_1                     :  2;
	unsigned int section_length                 : 12;
	unsigned int transport_stream_id            : 16;  
	unsigned int reserved_2                     :  2;
	unsigned int version_number                 :  5;
	unsigned int current_next_indicator         :  1;
	unsigned int section_number                 :  8;
	unsigned int last_section_number            :  8;
};

struct pat_field_item {
	unsigned int program_number         : 16;
	unsigned int reserved               :  3;
	unsigned int program_map_pid        : 13;
} ;

struct pat_hdr {
	unsigned int transport_stream_id        ;
	unsigned int version_number             ;
	unsigned int num_items                  ;
	struct       pat_item  *pat_item_ptr    ;
} ;

struct pat_item {
	unsigned int program_number             ;
	unsigned int program_map_pid            ;
	struct       queue_hdr  *pmt_q_hdr_ptr  ;
} ;

struct buf_hdr {
	unsigned short valid                    ; /* 0-no, 1-yes, 2-building */
							/* 3-wait for sect */
	unsigned short pid                      ;
	struct buf_hdr *q_ptr                   ;
	unsigned short sect_bad_missing         ;
	unsigned short ver_num                  ;
	unsigned int   sect_bytes_left          ;
	unsigned int   bytes_checked            ;
	unsigned int   byte_loc                 ;
	unsigned int   buf_size                 ;
	char           *buf_ptr                 ;
} ;

struct carousel{
	struct carousel  *nxt_carousel          ; /* ptr to next carousel */
	struct carousel  *bck_carousel          ; /* ptr to last carousel */
	unsigned short   assoc_tag              ;
	unsigned short   dii_pid                ;
	unsigned int     transaction_id         ;
	unsigned int     carousel_id            ;
	unsigned int     num_of_diis            ;
	unsigned int     private_len            ;
	char             *private_data          ;
	struct dii_level *dii_level_ptr         ; /* ptr to 1st dii struct */

} ;


struct dii_level{
	unsigned int     transaction_id    ;
	unsigned int     blk_size          ; /* bytes */
	unsigned int     grp_size          ; /* bytes */
	unsigned short   num_of_mods       ;
	unsigned short   pos               ; /* link cancatination position */
	unsigned int     private_len       ;
	char             *private_data     ;
	struct data_mod  *mod_array        ; /* ptr to 1st struct of a
					       contiguous allocation */
} ;

struct data_mod {
	unsigned short  module_id          ;
	unsigned short  mod_ver            ;
	unsigned short  pos                ; /* link cancatination position */
	unsigned int    nxt_module_id      ; /* next module id value        */
	unsigned int    mod_size           ; /* bytes */
	unsigned int    buf_size           ;
	unsigned int    num_of_blks        ;
	unsigned int    mod_not_done       ;
	char           *buffer             ; /* ptr to Module data buffer */
	struct blk_hdr *blk_hdr_ptr        ; /* ptr to 1st struct of a
						    contiguous allocation */
};

struct blk_hdr
{
	unsigned char   PTS_flag                ;
	unsigned char   valid_flag              ;
	unsigned short  PTS_hi                  ;
	unsigned int    PTS_low                 ;
};
/* --------------------------------------------------------------------- */
